;# $Id: hook.pl,v 2.9.1.1 92/08/26 13:14:05 ram Exp $
;#
;#  Copyright (c) 1992, Raphael Manfredi
;#
;#  You may redistribute only under the terms of the GNU General Public
;#  Licence as specified in the README file that comes with dist.
;#
;# $Log:	hook.pl,v $
;# Revision 2.9.1.1  92/08/26  13:14:05  ram
;# patch8: created
;# 
;# 
;# A mail hook (in the mailagent terminology) is an external file which
;# transparently influences some of the mailagent actions by injecting user-
;# defined actions at some well-defined places. Currently, the only hooks
;# available are executable folders, activated via the SAVE, STORE, and LEAVE
;# commands.
;#
;# The hook_type function parses the top of the hook file, looking for magic
;# token which will give hints regarding the type of the hook. Then the
;# corresponding hook function will be called with the file name where the mail
;# is stored given as first argument (an empty string meaning the mail is to be
;# fetched from stdin), the second argument being the hook file name.
;#
;# Five types of hooks are currently supported:
;#   - Simple program: the mail is simply fed to the standard input of the
;#     program. The exit status is propagated to the mailagent.
;#   - Rule file: the mail is to be re-analyzed according to the new rules
;#     held in the hook file. A new mailagent is called with the -r option.
;#   - Audit script: This is a perl script. Following the spirit of Martin
;#     Streicher's audit.pl package, some special variables are magically set
;#     prior to the invocation of script within the special mailhook package,
;#     in which the script is compiled.
;#   - Deliver script: Same as an audit script, excepted that the output of the
;#     script is monitored and taken as mailagent commands, which will then
;#     be executed on the original message upon completion of the script.
;#   - Perl script: This is an audit script with full access to the mailagent
;#     primitives for filtering (same as the ones provided with a PERL command).
;#
#
# Mailhook handling
#

package hook;

# Hooks constants definitions
sub main'init_hooks {
	$HOOK_UNKNOWN = "hook'unknown";		# Hook type was not recognized
	$HOOK_PROGRAM = "hook'program";		# Hook is a filter program
	$HOOK_AUDIT = "hook'audit";			# Hook is an audit-like script
	$HOOK_DELIVER = "hook'deliver";		# Hook is a deliver-like script
	$HOOK_RULES = "hook'rules";			# Hook is a rule file
	$HOOK_PERL = "hook'perl";			# Hook is a perl script
}

# Determine the nature of the hook. The top 128 bytes are scanned for a magic
# number starting with #: and followed by some words. The type of the hook
# is determined by the first word (case insensitively).
sub main'hook_type {
	local($file) = @_;			# Name of hook file
	-f "$file" || return $HOOK_UNKNOWN;		
	-x _ || return $HOOK_UNKNOWN;
	open(HOOK, $file) || return $HOOK_PROGRAM;
	local($hook) = ' ' x 128;	# Consider only top 128 bytes
	sysread(HOOK, $hook, 128);
	close(HOOK);
	local($name) = $hook =~ /^#:\s*(\w+)/;
	$name =~ tr/A-Z/a-z/;
	return $HOOK_AUDIT if $name eq 'audit';
	return $HOOK_DELIVER if $name eq 'deliver';
	return $HOOK_RULES if $name eq 'rules';
	return $HOOK_PERL if $name eq 'perl';
	$HOOK_PROGRAM;				# No magic token found
}

#
# Hook functions
#

# The hook file is not valid
sub unknown {
	local($mail, $hook) = @_;
	&'fatal("$hook is not a hook file");
}

# Mail is to be piped to the hook program (on stdin)
sub program {
	local($mail, $hook) = @_;
	&'add_log("hook is a plain program") if $'loglvl > 17;
	if ($mail ne '') {
		unless (open(MAIL, $mail)) {
			&'add_log("ERROR cannot open $mail: $!") if $'loglvl;
			&'fatal("cannot feed $hook");
		}
		# Child gets its mail from stdin anyway, so dup MAIL on STDIN
		unless (open(STDIN, '<&MAIL')) {
			&'add_log("ERROR cannot dup mail on stdin: $!") if $'loglvl;
			&'fatal("cannot feed $hook");
		}
	}
	exec $hook;		# Propagate exit status to parent (mailagent)
	# Exec system call failed
	&'add_log("ERROR cannot exec $hook: $!") if $'loglvl;
	&'fatal("cannot run $hook");
}

# Mail is to be filetered with rules from hook file
sub rules {
	local($mail, $hook) = @_;
	&'add_log("hook contains mailagent rules") if $'loglvl > 17;
	exec "perl -S mailagent -r $hook $mail";
	# Exec system call failed
	&'add_log("ERROR exec failed: $!") if $'loglvl;
	&'fatal("cannot run mailagent");
}

# Mail is to be filtered through a perl script
sub perl {
	local($mail, $hook) = @_;
	&'add_log("hook is a perl script") if $'loglvl > 17;
	exec "perl -S mailagent -e 'PERL $hook' $mail";
	# Exec system call failed
	&'add_log("ERROR exec failed: $!") if $'loglvl;
	&'fatal("cannot run mailagent");
}

# Hook is an audit script. Set up a suitable environment and execute the
# script without forking any new process. To avoid name clashes, the script
# is compiled in a dedicated 'mailhook' package and executed.
sub audit {
	local($mail, $hook) = @_;
	&'add_log("hook is an audit script") if $'loglvl > 17;
	&'parse_mail($mail);		# Fill in %Header
	&initialize;				# Initialize special variables
	&run($hook);				# Load hook and run it
}

# A delivery script is about the same as an audit script, except that the
# output on stdout is monitored and understood as mailagent commands to be
# executed via a '-e' switch upon successful return.
sub deliver {
	local($mail, $hook) = @_;
	local($pid);
	&'add_log("hook is a deliver script") if $'loglvl > 17;
	# Fork and let the child do all the work. The parent simply captures the
	# output from child's stdout.
	&'parse_mail($mail);		# Fill in %Header
	$pid = open(HOOK, "-|");	# Implicit fork
	unless (defined $pid) {
		&'add_log("ERROR cannot fork: $!") if $'loglvl;
		&'fatal("cannot deliver to hook");
	}
	if (0 == $pid) {			# Child process
		&initialize;			# Initialize special variables
		&run($hook);			# Load hook and run it
		exit 0;					# Everything went well
	}
	local($/) = undef;			# We wish to slurp the whole output
	local($output) = <HOOK>;
	close HOOK;					# An implicit wait -- status put in $?
	unless (0 == $?) {
		&'add_log("ERROR hook script failed") if $'loglvl;
		&'fatal("non-zero exit status") unless $output;
		&'fatal("commands ignored");
	}
	if ($output eq '') {
		&'add_log("WARNING no commands from delivery hook") if $'loglvl > 5;
	} else {
		local($*) = 1;
		$output =~ s/\\/\\\\/g;				# Protect backslashes
		$output =~ s/([ "'`;<>&|])/\\$1/g;	# Protect spaces, separators, etc...
		$pid = open(MAILAGENT, "|perl -S mailagent -e $output");
		unless (defined $pid) {
			&'add_log("ERROR cannot fork: $!") if $'loglvl;
			&'fatal("cannot run mailagent");
		}
		print MAILAGENT $'Header{'All'};	# Pipe mail to mailagent
		close MAILAGENT;					# Implicit wait
		&'fatal("mailagent failed") if $?;
	}
}

# Log hook operation before it happens, as we may well exec() another program.
sub main'hooking {
	local($file, $hook, $type) = @_;
	$file =~ s|.*/(.*)|$1|;			# Keep only base name
	$file = '<stdin>' if $file eq '';
	$type =~ s/^hook'//;
	$hook =~ s/^$cf'home/~/;
	&'add_log("HOOKING [$file] to $hook ($type)") if $'loglvl > 4;
}

package main;

